package ink.wgink.module.file.service.impl;

import ink.wgink.common.base.DefaultBaseService;
import ink.wgink.exceptions.FileException;
import ink.wgink.exceptions.base.SystemException;
import ink.wgink.module.file.dao.IFileDao;
import ink.wgink.module.file.enums.UploadTypeEnum;
import ink.wgink.module.file.pojo.dtos.FileDTO;
import ink.wgink.module.file.pojo.dtos.FileInfoDTO;
import ink.wgink.module.file.pojo.vos.FileVO;
import ink.wgink.module.file.service.IDefaultFileService;
import ink.wgink.module.file.service.IFileService;
import ink.wgink.module.file.service.IMinIoFileService;
import ink.wgink.pojo.ListPage;
import ink.wgink.pojo.pos.FilePO;
import ink.wgink.pojo.result.SuccessResultList;
import ink.wgink.properties.FileMinIoProperties;
import ink.wgink.properties.FileProperties;
import ink.wgink.util.ArrayListUtil;
import ink.wgink.util.UUIDUtil;
import ink.wgink.util.map.HashMapUtil;
import ink.wgink.util.request.StaticResourceRequestUtil;
import io.minio.*;
import io.minio.errors.*;
import org.apache.commons.lang3.StringUtils;
import org.joda.time.DateTime;
import org.joda.time.DateTimeZone;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Pageable;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;

import javax.annotation.PostConstruct;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URLEncoder;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

/**
 * @ClassName: MinIOServiceImpl
 * @Description: minIO
 * @Author: wanggeng
 * @Date: 2021/10/17 10:16 上午
 * @Version: 1.0
 */
@Service
public class MinIoFileServiceImpl extends DefaultBaseService implements IMinIoFileService {

    public static final String UPLOAD_FILE_COLLECTION_NAME = "uploadFiles";
    @Autowired
    private FileProperties fileProperties;
    @Autowired
    private IFileDao fileDao;
    @Autowired
    private IDefaultFileService defaultFileService;
    @Autowired(required = false)
    private MongoTemplate mongoTemplate;
    private FileMinIoProperties minIoProperties;
    private MinioClient minioClient;

    @PostConstruct
    public void init() {
        if (!fileProperties.getUseMinIo()) {
            return;
        }
        minIoProperties = fileProperties.getMinIo();
        minioClient = MinioClient.builder()
                .endpoint(minIoProperties.getEndpoint())
                .credentials(minIoProperties.getAccessKey(), minIoProperties.getSecretKey())
                .build();
    }

    @Override
    public String saveFileByUserId(String userId, FileVO fileVO) {
        String fileId = defaultFileService.saveFileByUserId(userId, fileVO);
        if (mongoTemplate != null) {
            Map<String, Object> params = HashMapUtil.beanToMap(fileVO);
            params.put("fileId", fileId);
            params.put("isBack", 0);
            setSaveInfoByUserId(params, userId);
            mongoTemplate.insert(params);
        }
        return fileId;
    }

    @Override
    public void remove(List<String> ids) {
        if (ids.isEmpty()) {
            return;
        }
        defaultFileService.remove(ids);
        // 删除缓存
        if (mongoTemplate != null) {
            mongoTemplate.remove(Query.query(Criteria.where("fileId").in(ids)), UPLOAD_FILE_COLLECTION_NAME);
        }
    }

    @Override
    public void deleteFile(List<FilePO> filePOs) {
        // 删除数据
        if (filePOs.isEmpty()) {
            return;
        }
        List<String> fileIds = ArrayListUtil.listBeanStringIdValue(filePOs, "fileId", FilePO.class);
        defaultFileService.delete(fileIds);
        if (mongoTemplate != null) {
            mongoTemplate.remove(Query.query(Criteria.where("fileId").in(fileIds)));
        }
    }

    @Override
    public void uploadFile(String token, MultipartFile uploadFile, UploadTypeEnum uploadTypeEnum, Map<String, Object> params) {
        uploadFile(token, uploadFile, uploadTypeEnum, params, false);
    }

    @Override
    public void uploadFileByUserId(String userId, MultipartFile uploadFile, UploadTypeEnum uploadTypeEnum, Map<String, Object> params) {
        uploadFile(userId, uploadFile, uploadTypeEnum, params, true);
    }

    @Override
    public boolean downLoadFile(HttpServletRequest request, HttpServletResponse response, Map<String, Object> params, boolean canRange) {
        FilePO filePO = null;
        if (mongoTemplate != null) {
            filePO = mongoTemplate.findOne(Query.query(Criteria.where("fileId").is(params.get("fileId").toString())), FilePO.class, UPLOAD_FILE_COLLECTION_NAME);
        }
        if (filePO == null) {
            filePO = fileDao.getPO(params);
        }
        if (filePO == null) {
            throw new FileException("文件不存在");
        }
        boolean isOpen = Boolean.valueOf(params.get("isOpen").toString());
        return downLoadFile(request, response, filePO, isOpen, canRange);
    }

    /**
     * 上传文件
     *
     * @param tokenOrUserId
     * @param uploadFile
     * @param uploadTypeEnum
     * @param params
     * @param useUserId      是否使用外来userId
     */
    private void uploadFile(String tokenOrUserId, MultipartFile uploadFile, UploadTypeEnum uploadTypeEnum, Map<String, Object> params, boolean useUserId) {
        String fileId = UUIDUtil.getUUID();
        params.put("fileId", fileId);

        String fileName = uploadFile.getOriginalFilename();
        // 文件大小
        long fileSize = uploadFile.getSize();
        // 文件类型
        String fileType = getFileType(fileName);
        String uploadFileName = fileId + "." + fileType;
        String bucketName = getBucketName(uploadTypeEnum);
        // 上传文件
        upload(uploadFileName, fileSize, bucketName, uploadFile);
        if (useUserId) {
            saveFileByUserId(tokenOrUserId, fileId, fileName, fileType, fileSize, bucketName);
        } else {
            saveFile(tokenOrUserId, fileId, fileName, fileType, fileSize, bucketName);
        }
    }

    @Override
    public boolean downLoadFile(HttpServletRequest request, HttpServletResponse response, FilePO filePO, boolean isOpen, boolean canRange) {
        boolean isDownloadComplete = false;
        // minIo中文件名
        String objectName = filePO.getFileId() + "." + filePO.getFileType();
        InputStream inputStream = null;
        try (OutputStream outputStream = response.getOutputStream();) {
            response.setHeader("Content-Length", filePO.getFileSize());
            response.setContentType(StaticResourceRequestUtil.getContentType(filePO.getFileType()));
            if (!isOpen) {
                // 下载
                response.setHeader("Content-Disposition", "attachment;fileName=" + URLEncoder.encode(filePO.getFileName(), "UTF-8"));
                GetObjectArgs getObjectArgs = GetObjectArgs.builder().bucket(filePO.getFileUrl()).object(objectName).build();
                inputStream = minioClient.getObject(getObjectArgs);
            } else {
                // 直接打开
                response.setHeader("Content-Disposition", "inline;fileName=" + URLEncoder.encode(filePO.getFileName(), "UTF-8"));
                // 如果是图片资源，开启缓存
                if (isImageFile(filePO.getFileType())) {
                    // 如果存在校验修改时间且未做修改，返回304，使用本地资源
                    String ifModifiedSince = request.getHeader("If-Modified-Since");
                    if (StringUtils.isNotBlank(ifModifiedSince) && StringUtils.equalsIgnoreCase(ifModifiedSince, filePO.getGmtModified())) {
                        response.setStatus(HttpServletResponse.SC_NOT_MODIFIED);
                        return true;
                    }
                    // 缓存有效时间为7天
                    response.setHeader("Expires", DateTime.now().plusDays(7).toDateTime(DateTimeZone.forID("GMT")).toString());
                    // 一小时之内不发送新请求
                    response.setHeader("max-age", "3600");
                    response.setHeader("Last-Modified", filePO.getGmtModified());
                }
                String rangeString = null;
                if (canRange && request != null) {
                    rangeString = request.getHeader("Range");
                    LOG.debug("range: {}", rangeString);
                }
                long contentLength = Long.valueOf(filePO.getFileSize());
                long startRange = 0;
                long endRange = contentLength - 1;
                if (!StringUtils.isBlank(rangeString)) {
                    response.setContentType("multipart/byteranges");
                    String[] rangeArray = rangeString.substring(rangeString.indexOf("=") + 1).split("-");
                    startRange = Long.valueOf(rangeArray[0]);
                    if (rangeArray.length > 1) {
                        endRange = Long.valueOf(rangeArray[1]);
                    }
                    setRangeHeader(startRange, endRange, response, filePO.getFileId(), contentLength);
                    inputStream = minioClient.getObject(GetObjectArgs.builder().bucket(filePO.getFileUrl()).object(objectName).offset(startRange).length(endRange - startRange).build());
                } else {
                    inputStream = minioClient.getObject(GetObjectArgs.builder().bucket(filePO.getFileUrl()).object(objectName).build());
                }
                if (endRange == contentLength - 1) {
                    isDownloadComplete = true;
                }
            }
            byte[] readBuf = new byte[IFileService.INPUT_STREAM_SIZE];
            for (int length = 0; (length = inputStream.read(readBuf)) > 0; ) {
                outputStream.write(readBuf, 0, length);
            }
            outputStream.flush();
            inputStream.close();
        } catch (ServerException | IOException | ErrorResponseException | InsufficientDataException | InternalException | InvalidKeyException | InvalidResponseException | NoSuchAlgorithmException | XmlParserException e) {
            throw new FileException("文件输出异常", e);
        } finally {
            if (inputStream != null) {
                try {
                    inputStream.close();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }
        return isDownloadComplete;
    }

    @Override
    public SuccessResultList<List<FileInfoDTO>> listPageInfo(ListPage page) {
        SuccessResultList<List<FileInfoDTO>> successResultList = null;
        if (mongoTemplate != null) {
            Query query = new Query();
            String keywords = getKeywords(page.getParams());
            if (!StringUtils.isBlank(keywords)) {
                Pattern pattern = Pattern.compile("^.*" + keywords + ".*$", Pattern.CASE_INSENSITIVE);
                query.addCriteria(Criteria.where("fileName").regex(pattern));
            }
            long total = mongoTemplate.count(query, UPLOAD_FILE_COLLECTION_NAME);
            query.with(Pageable.ofSize(page.getRows()).withPage(page.getPage() - 1));
            List<FileInfoDTO> dataDTOs = mongoTemplate.find(query, FileInfoDTO.class, UPLOAD_FILE_COLLECTION_NAME);
            successResultList = new SuccessResultList<>(dataDTOs, page.getPage(), total);
        }
        if (successResultList != null && !successResultList.getRows().isEmpty()) {
            return successResultList;
        }
        return defaultFileService.listPageInfo(page);
    }

    @Override
    public List<FileDTO> list(List<String> idList) {
        List<FileDTO> fileDTOs = new ArrayList<>();
        if (mongoTemplate != null) {
            return mongoTemplate.find(Query.query(Criteria.where("fileId").in(idList)), FileDTO.class, UPLOAD_FILE_COLLECTION_NAME);
        }
        if (!fileDTOs.isEmpty()) {
            return fileDTOs;
        }
        return defaultFileService.list(idList);
    }

    @Override
    public List<FilePO> listPO(List<String> idList) {
        List<FilePO> filePOs = new ArrayList<>();
        if (mongoTemplate != null) {
            filePOs = mongoTemplate.find(Query.query(Criteria.where("fileId").in(idList)), FilePO.class, UPLOAD_FILE_COLLECTION_NAME);
        }
        if (!filePOs.isEmpty()) {
            return filePOs;
        }
        return defaultFileService.listPO(idList);
    }

    @Override
    public FilePO getPO(String fileId) {
        FilePO filePO = null;
        if (mongoTemplate != null) {
            filePO = mongoTemplate.findOne(Query.query(Criteria.where("fileId").is(fileId)), FilePO.class, UPLOAD_FILE_COLLECTION_NAME);
        }
        if (filePO != null) {
            return filePO;
        }
        return defaultFileService.getPO(fileId);
    }

    @Override
    public List<FileDTO> listByUrl(String fileUrl) {
        List<FileDTO> fileDTOs = new ArrayList<>();
        if (mongoTemplate != null) {
            Pattern pattern = Pattern.compile("^.*" + fileUrl + ".*$", Pattern.CASE_INSENSITIVE);
            fileDTOs = mongoTemplate.find(Query.query(Criteria.where("fileUrl").regex(pattern)), FileDTO.class, UPLOAD_FILE_COLLECTION_NAME);
        }
        if (!fileDTOs.isEmpty()) {
            return fileDTOs;
        }
        return defaultFileService.listByUrl(fileUrl);
    }


    /**
     * 上传文件到minIo
     *
     * @param uploadFileName
     * @param fileSize
     * @param bucketName
     * @param uploadFile
     */
    private void upload(String uploadFileName, long fileSize, String bucketName, MultipartFile uploadFile) {
        try (InputStream inputStream = uploadFile.getInputStream()) {
            // 创建桶
            if (!minioClient.bucketExists(BucketExistsArgs.builder().bucket(bucketName).build())) {
                minioClient.makeBucket(MakeBucketArgs.builder().bucket(bucketName).build());
            }
            minioClient.putObject(PutObjectArgs.builder().bucket(bucketName).object(uploadFileName).stream(inputStream, fileSize, -1).contentType(uploadFile.getContentType()).build());
        } catch (MinioException | InvalidKeyException | IOException | NoSuchAlgorithmException e) {
            throw new SystemException(e);
        }
    }

    /**
     * 保存文件
     *
     * @param token
     * @param fileId
     * @param fileName
     * @param fileType
     * @param fileSize
     * @param bucketName 桶名称
     */
    private void saveFile(String token, String fileId, String fileName, String fileType, long fileSize, String bucketName) {
        Map<String, Object> params = getFileParams(fileId, fileName, fileType, fileSize, bucketName);
        if (StringUtils.isBlank(token)) {
            setSaveInfo(params);
        } else {
            setAppSaveInfo(token, params);
        }
        fileDao.save(params);
        if (mongoTemplate != null) {
            mongoTemplate.insert(params, UPLOAD_FILE_COLLECTION_NAME);
        }
    }

    /**
     * 保存用户
     *
     * @param userId     用户ID
     * @param fileId
     * @param fileName
     * @param fileType
     * @param fileSize
     * @param bucketName
     */
    private void saveFileByUserId(String userId, String fileId, String fileName, String fileType, long fileSize, String bucketName) {
        Map<String, Object> params = getFileParams(fileId, fileName, fileType, fileSize, bucketName);
        setSaveInfoByUserId(params, userId);
        fileDao.save(params);
        if (mongoTemplate != null) {
            mongoTemplate.insert(params, UPLOAD_FILE_COLLECTION_NAME);
        }
    }

    private Map<String, Object> getFileParams(String fileId, String fileName, String fileType, long fileSize, String bucketName) {
        Map<String, Object> params = getHashMap(2);
        params.put("fileId", fileId);
        params.put("fileName", fileName);
        params.put("filePath", null);
        params.put("fileUrl", bucketName);
        params.put("fileType", fileType);
        params.put("fileSize", fileSize);
        params.put("fileSummary", "minio");
        params.put("isBack", 1);
        return params;
    }

    /**
     * 获取文件类型
     *
     * @param fileName
     * @return
     */
    private String getFileType(String fileName) {
        String[] names = fileName.split("\\.");
        if (names != null) {
            return names[names.length - 1].toLowerCase();
        }
        return "";
    }

    /**
     * 获取桶文件名
     *
     * @param uploadTypeEnum 上传类型
     * @return
     */
    private String getBucketName(UploadTypeEnum uploadTypeEnum) {
        if (uploadTypeEnum.getValue() == UploadTypeEnum.IMAGE.getValue()) {
            return "images";
        } else if (uploadTypeEnum.getValue() == UploadTypeEnum.VIDEO.getValue()) {
            return "videos";
        } else if (uploadTypeEnum.getValue() == UploadTypeEnum.AUDIO.getValue()) {
            return "audios";
        }
        return "files";
    }

    /**
     * 是否是图片
     *
     * @param fileType
     * @return
     */
    private boolean isImageFile(String fileType) {
        if (StringUtils.equalsIgnoreCase("gif", fileType)) {
            return false;
        }
        String imageTypes = fileProperties.getImageTypes();
        for (String imageType : imageTypes.split(",")) {
            if (StringUtils.equalsIgnoreCase(fileType, imageType)) {
                return true;
            }
        }
        return false;
    }

    /**
     * 处理视频流问题
     *
     * @param startRange
     * @param endRange
     * @param response
     * @param fileId
     * @param contentLength
     */
    private void setRangeHeader(long startRange, long endRange, HttpServletResponse response, String fileId, long contentLength) {
        // 这里不设置，会出现第一次加载很慢的情况
        response.setHeader("Content-Length", String.valueOf(endRange - startRange + 1));
        response.setHeader("Content-Range", String.format("bytes %d-%d/%d", startRange, endRange, contentLength));
        response.setHeader("Accept-Ranges", "bytes");
        response.setHeader("Etag", fileId);
        response.setStatus(206);
    }
}
